"""CORSIKA configuration."""

import logging
from pathlib import Path

import numpy as np
from astropy import units as u

from simtools.corsika.primary_particle import PrimaryParticle
from simtools.io import eventio_handler, io_handler
from simtools.model.model_parameter import ModelParameter
from simtools.utils import general as gen


class CorsikaConfig:
    """
    Configuration for the CORSIKA air shower simulation software.

    Follows closely the CORSIKA definitions and output format (see CORSIKA manual).

    The configuration is set as a dict corresponding to the command line configuration groups
    (especially simulation_software, simulation configuration, simulation parameters).

    Parameters
    ----------
    array_model : ArrayModel
        Array model.
    args_dict : dict
        Configuration dictionary.
    db_config : dict
        MongoDB configuration.
    label : str
        Instance label.
    dummy_simulations : bool
        If True, the configuration is generated for dummy simulations
        (e.g., sim_telarray requires for some run modes a valid CORSIKA input file).
    """

    def __init__(self, array_model, args_dict, db_config=None, label=None, dummy_simulations=False):
        """Initialize CorsikaConfig."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init CorsikaConfig")

        self.label = label
        self.shower_events = self.mc_events = None
        self.zenith_angle = self.azimuth_angle = None
        self.curved_atmosphere_min_zenith_angle = None
        self._run_number = None
        self.config_file_path = None
        self.primary_particle = args_dict  # see setter for primary_particle
        self.use_curved_atmosphere = args_dict  # see setter for use_curved_atmosphere
        self.dummy_simulations = dummy_simulations

        self.io_handler = io_handler.IOHandler()
        self.array_model = array_model
        self.config = self._fill_corsika_configuration(args_dict, db_config)
        self._initialize_from_config(args_dict)
        self.is_file_updated = False

    @property
    def primary_particle(self):
        """Primary particle."""
        return self._primary_particle

    @primary_particle.setter
    def primary_particle(self, args):
        """
        Set primary particle from input dictionary or CORSIKA 7 particle ID.

        This is to make sure that when setting the primary particle,
        we get the full PrimaryParticle object expected.

        Parameters
        ----------
        args: dict, corsika particle ID, or None
            Configuration dictionary
        """
        if (
            isinstance(args, dict)
            and args.get("primary_id_type") is not None
            and args.get("primary") is not None
        ):
            self._primary_particle = PrimaryParticle(
                particle_id_type=args.get("primary_id_type"), particle_id=args.get("primary")
            )
        elif isinstance(args, int):
            self._primary_particle = PrimaryParticle(
                particle_id_type="corsika7_id", particle_id=args
            )
        else:
            self._primary_particle = PrimaryParticle()

    @property
    def use_curved_atmosphere(self):
        """Check if zenith angle condition for curved atmosphere usage for CORSIKA is met."""
        return self._use_curved_atmosphere

    @use_curved_atmosphere.setter
    def use_curved_atmosphere(self, args):
        """Check if zenith angle condition for curved atmosphere usage for CORSIKA is met."""
        self._use_curved_atmosphere = False
        if isinstance(args, bool):
            self._use_curved_atmosphere = args
        elif isinstance(args, dict):
            try:
                self._use_curved_atmosphere = (
                    args.get("zenith_angle", 0.0 * u.deg).to("deg").value
                    > args["curved_atmosphere_min_zenith_angle"].to("deg").value
                )
            except KeyError:
                self._use_curved_atmosphere = False

    def _fill_corsika_configuration(self, args_dict, db_config=None):
        """
        Fill CORSIKA configuration.

        Dictionary keys are CORSIKA parameter names. Values are converted to
        CORSIKA-consistent units.

        Parameters
        ----------
        args_dict : dict
            Configuration dictionary.
        db_config: dict
            Database configuration.

        Returns
        -------
        dict
            Dictionary with CORSIKA parameters.
        """
        if args_dict is None:
            return {}

        config = {}
        if self.dummy_simulations:
            config["USER_INPUT"] = self._corsika_configuration_for_dummy_simulations(args_dict)
        elif args_dict.get("corsika_file", None) is not None:
            config["USER_INPUT"] = self._corsika_configuration_from_corsika_file(
                args_dict["corsika_file"]
            )
        else:
            config["USER_INPUT"] = self._corsika_configuration_from_user_input(args_dict)

        config.update(
            self._fill_corsika_configuration_from_db(
                gen.ensure_iterable(args_dict.get("model_version")), db_config
            )
        )
        return config

    def _fill_corsika_configuration_from_db(self, model_versions, db_config):
        """Fill CORSIKA configuration from database."""
        config = {}
        if db_config is None:  # all following parameter require DB
            return config

        # For multiple model versions, check that CORSIKA parameters are identical
        self.assert_corsika_configurations_match(model_versions, db_config=db_config)
        model_version = model_versions[0]

        self._logger.debug(f"Using model version {model_version} for CORSIKA parameters from DB")
        db_model_parameters = ModelParameter(db_config=db_config, model_version=model_version)
        parameters_from_db = db_model_parameters.get_simulation_software_parameters("corsika")

        config["INTERACTION_FLAGS"] = self._corsika_configuration_interaction_flags(
            parameters_from_db
        )
        config["CHERENKOV_EMISSION_PARAMETERS"] = self._corsika_configuration_cherenkov_parameters(
            parameters_from_db
        )
        config["DEBUGGING_OUTPUT_PARAMETERS"] = self._corsika_configuration_debugging_parameters()
        config["IACT_PARAMETERS"] = self._corsika_configuration_iact_parameters(parameters_from_db)
        return config

    def _initialize_from_config(self, args_dict):
        """
        Initialize additional parameters either from command line args or from derived config.

        Takes into account that in the case of a given CORSIKA input file, some parameters are read
        from the file instead of the command line args.

        """
        self.primary_particle = int(self.config.get("USER_INPUT", {}).get("PRMPAR", [1])[0])
        self.shower_events = int(self.config.get("USER_INPUT", {}).get("NSHOW", [0])[0])
        self.mc_events = int(
            self.shower_events * self.config.get("USER_INPUT", {}).get("CSCAT", [1])[0]
        )

        if args_dict.get("corsika_file", None) is not None:
            azimuth = self._rotate_azimuth_by_180deg(
                0.5 * (self.config["USER_INPUT"]["PHIP"][0] + self.config["USER_INPUT"]["PHIP"][1]),
                invert_operation=True,
            )
            zenith = 0.5 * (
                self.config["USER_INPUT"]["THETAP"][0] + self.config["USER_INPUT"]["THETAP"][1]
            )
        else:
            azimuth = args_dict.get("azimuth_angle", 0.0 * u.deg).to("deg").value
            zenith = args_dict.get("zenith_angle", 20.0 * u.deg).to("deg").value

        self.azimuth_angle = round(azimuth)
        self.zenith_angle = round(zenith)

        self.curved_atmosphere_min_zenith_angle = (
            args_dict.get("curved_atmosphere_min_zenith_angle", 90.0 * u.deg).to("deg").value
        )

    def assert_corsika_configurations_match(self, model_versions, db_config=None):
        """
        Assert that CORSIKA configurations match across all model versions.

        Parameters
        ----------
        model_versions : list
            List of model versions to check.
        db_config : dict, optional
            Database configuration.

        Raises
        ------
        InvalidCorsikaInputError
            If CORSIKA parameters are not identical across all model versions.
        """
        if len(model_versions) < 2:
            return

        parameters_from_db_list = []

        # Get parameters for all model versions
        for model_version in model_versions:
            db_model_parameters = ModelParameter(db_config=db_config, model_version=model_version)
            parameters_from_db_list.append(
                db_model_parameters.get_simulation_software_parameters("corsika")
            )

        # Parameters that can differ between model versions (e.g., i/o buffer size)
        skip_parameters = ["corsika_iact_io_buffer", "corsika_iact_split_auto"]

        # Check if all parameters match
        for i in range(len(parameters_from_db_list) - 1):
            for key in parameters_from_db_list[i]:
                if key in skip_parameters:
                    continue

                current_value = parameters_from_db_list[i][key]["value"]
                next_value = parameters_from_db_list[i + 1][key]["value"]

                if current_value != next_value:
                    self._logger.warning(
                        f"Parameter '{key}' mismatch between model versions:\n"
                        f"  {model_versions[i]}: {current_value}\n"
                        f"  {model_versions[i + 1]}: {next_value}"
                    )
                    raise ValueError(
                        f"CORSIKA parameter '{key}' differs between model versions "
                        f"{model_versions[i]} and {model_versions[i + 1]}. "
                        f"Values are {current_value} and {next_value} respectively."
                    )

    def _corsika_configuration_for_dummy_simulations(self, args_dict):
        """
        Return CORSIKA configuration for dummy simulations.

        Settings are such that that the simulations are fast
        and none (or not many) Cherenkov photons are generated.

        Returns
        -------
        dict
            Dictionary with CORSIKA parameters for dummy simulations.
        """
        theta, phi = self._get_corsika_theta_phi(args_dict)
        return {
            "EVTNR": [1],
            "NSHOW": [1],
            "PRMPAR": [1],  # CORSIKA ID 1 for primary gamma
            "ESLOPE": [-2.0],
            "ERANGE": [0.1, 0.1],
            "THETAP": [theta, theta],
            "PHIP": [phi, phi],
            "VIEWCONE": [0.0, 0.0],
            "CSCAT": [1, 0.0, 10.0],
        }

    def _corsika_configuration_from_corsika_file(self, corsika_input_file):
        """
        Get CORSIKA configuration run header of provided input files.

        Reads configuration from the run and event headers from the CORSIKA input file
        (unfortunately quite fine tuned to the pycorsikaio run and event
        header implementation).

        Parameters
        ----------
        corsika_input_file : str, path
            Path to the CORSIKA input file.

        Returns
        -------
        dict
            Dictionary with CORSIKA parameters from input file.
        """
        run_header, event_header = eventio_handler.get_corsika_run_and_event_headers(
            corsika_input_file
        )
        self._logger.debug(f"CORSIKA run header from {corsika_input_file}")

        def to_float32(value):
            """Convert value to numpy float32."""
            return np.float32(value) if value is not None else 0.0

        def to_int32(value):
            """Convert value to numpy int32."""
            return np.int32(value) if value is not None else 0

        if run_header["n_observation_levels"] > 0:
            self._check_altitude_and_site(run_header["observation_height"][0])

        return {
            "EVTNR": [to_int32(event_header["event_number"])],
            "NSHOW": [to_int32(run_header["n_showers"])],
            "PRMPAR": [to_int32(event_header["particle_id"])],
            "ESLOPE": [to_float32(run_header["energy_spectrum_slope"])],
            "ERANGE": [to_float32(run_header["energy_min"]), to_float32(run_header["energy_max"])],
            "THETAP": [
                to_float32(event_header["theta_min"]),
                to_float32(event_header["theta_max"]),
            ],
            "PHIP": [to_float32(event_header["phi_min"]), to_float32(event_header["phi_max"])],
            "VIEWCONE": [
                to_float32(event_header["viewcone_inner_angle"]),
                to_float32(event_header["viewcone_outer_angle"]),
            ],
            "CSCAT": [
                to_int32(event_header["n_reuse"]),
                to_float32(event_header["reuse_x"]),
                to_float32(event_header["reuse_y"]),
            ],
        }

    def _corsika_configuration_from_user_input(self, args_dict):
        """
        Get CORSIKA configuration from user input.

        Parameters
        ----------
        args_dict : dict
            Configuration dictionary.

        Returns
        -------
        dict
            Dictionary with CORSIKA parameters.
        """
        theta, phi = self._get_corsika_theta_phi(args_dict)
        return {
            "EVTNR": [args_dict["event_number_first_shower"]],
            "NSHOW": [args_dict["nshow"]],
            "PRMPAR": [self.primary_particle.corsika7_id],
            "ESLOPE": [args_dict["eslope"]],
            "ERANGE": [
                args_dict["energy_range"][0].to("GeV").value,
                args_dict["energy_range"][1].to("GeV").value,
            ],
            "THETAP": [theta, theta],
            "PHIP": [phi, phi],
            "VIEWCONE": [
                args_dict["view_cone"][0].to("deg").value,
                args_dict["view_cone"][1].to("deg").value,
            ],
            "CSCAT": [
                args_dict["core_scatter"][0],
                args_dict["core_scatter"][1].to("cm").value,
                0.0,
            ],
        }

    def _check_altitude_and_site(self, observation_height):
        """Check that observation height from CORSIKA file matches site model."""
        site_altitude = self.array_model.site_model.get_parameter_value("corsika_observation_level")
        if not np.isclose(observation_height / 1.0e2, site_altitude, atol=1.0):
            raise ValueError(
                "Observatory altitude does not match CORSIKA file observation height: "
                f"{site_altitude} m (site model) != {observation_height / 1.0e2} m (CORSIKA file)"
            )

    def _get_corsika_theta_phi(self, args_dict):
        """Get CORSIKA theta and phi angles from args_dict."""
        theta = args_dict.get("zenith_angle", 20.0 * u.deg).to("deg").value
        phi = self._rotate_azimuth_by_180deg(
            args_dict.get("azimuth_angle", 0.0 * u.deg).to("deg").value,
            correct_for_geomagnetic_field_alignment=args_dict.get(
                "correct_for_b_field_alignment", True
            ),
        )
        return theta, phi

    def _corsika_configuration_interaction_flags(self, parameters_from_db):
        """
        Return CORSIKA interaction flags / parameters.

        Parameters
        ----------
        parameters_from_db : dict
            CORSIKA parameters from the database.

        Returns
        -------
        interaction_parameters : dict
            Dictionary with CORSIKA interaction parameters.
        """
        parameters = {}
        parameters["FIXHEI"] = self._input_config_first_interaction_height(
            parameters_from_db["corsika_first_interaction_height"]
        )
        parameters["FIXCHI"] = [
            self._input_config_corsika_starting_grammage(
                parameters_from_db["corsika_starting_grammage"]
            )
        ]
        if not self.use_curved_atmosphere:
            parameters["TSTART"] = ["T"]
        parameters["ECUTS"] = self._input_config_corsika_particle_kinetic_energy_cutoff(
            parameters_from_db["corsika_particle_kinetic_energy_cutoff"]
        )
        parameters["MUADDI"] = ["F"]
        parameters["MUMULT"] = ["T"]
        parameters["LONGI"] = self._input_config_corsika_longitudinal_parameters(
            parameters_from_db["corsika_longitudinal_shower_development"]
        )
        parameters["MAXPRT"] = ["10"]
        parameters["ECTMAP"] = ["1.e6"]

        self._logger.debug(f"Interaction parameters: {parameters}")
        return parameters

    def _input_config_first_interaction_height(self, entry):
        """Return FIXHEI parameter CORSIKA format."""
        return [f"{entry['value'] * u.Unit(entry['unit']).to('cm'):.2f}", "0"]

    def _input_config_corsika_starting_grammage(self, entry):
        """Return FIXCHI parameter CORSIKA format."""
        value = self._get_starting_grammage_value(entry["value"])
        return f"{value * u.Unit(entry['unit']).to('g/cm2')}"

    def _get_starting_grammage_value(self, value_entry):
        """
        Get appropriate starting grammage value from entry values.

        Parameters
        ----------
        value_entry : float or list
            Value or list of grammage configurations

        Returns
        -------
        float
            Selected grammage value
        """
        if not isinstance(value_entry, list):
            return value_entry

        tel_types = {tel.design_model for tel in self.array_model.telescope_model.values()}
        particle = self.primary_particle.name
        matched_values = self._get_matching_grammage_values(value_entry, tel_types, particle)

        return min(matched_values) if matched_values else 0

    def _get_matching_grammage_values(self, configs, tel_types, particle):
        """Get list of matching grammage values for particle and telescope types."""
        matched = []
        defaults = []

        for config in configs:
            if config.get("instrument") is None or config.get("instrument") in tel_types:
                if config["primary_particle"] == particle:
                    matched.append(config["value"])
                elif config["primary_particle"] == "default":
                    defaults.append(config["value"])

        return matched if matched else defaults

    def _input_config_corsika_particle_kinetic_energy_cutoff(self, entry):
        """Return ECUTS parameter CORSIKA format."""
        e_cuts = entry["value"]
        return [
            f"{e_cuts[0] * u.Unit(entry['unit']).to('GeV')} "
            f"{e_cuts[1] * u.Unit(entry['unit']).to('GeV')} "
            f"{e_cuts[2] * u.Unit(entry['unit']).to('GeV')} "
            f"{e_cuts[3] * u.Unit(entry['unit']).to('GeV')}"
        ]

    def _input_config_corsika_longitudinal_parameters(self, entry):
        """Return LONGI parameter CORSIKA format."""
        return ["T", f"{entry['value'] * u.Unit(entry['unit']).to('g/cm2')}", "F", "F"]

    def _corsika_configuration_cherenkov_parameters(self, parameters_from_db):
        """
        Return CORSIKA Cherenkov emission parameters.

        Parameters
        ----------
        parameters_from_db : dict
            CORSIKA parameters from the database.

        Returns
        -------
        dict
            Dictionary with CORSIKA Cherenkov emission parameters.
        """
        parameters = {}
        parameters["CERSIZ"] = [parameters_from_db["corsika_cherenkov_photon_bunch_size"]["value"]]
        parameters["CERFIL"] = "0"
        parameters["CWAVLG"] = self._input_config_corsika_cherenkov_wavelength(
            parameters_from_db["corsika_cherenkov_photon_wavelength_range"]
        )
        self._logger.debug(f"Cherenkov parameters: {parameters}")
        return parameters

    def _input_config_corsika_cherenkov_wavelength(self, entry):
        """Return CWAVLG parameter CORSIKA format."""
        wavelength_range = entry["value"]
        return [
            f"{wavelength_range[0] * u.Unit(entry['unit']).to('nm')}",
            f"{wavelength_range[1] * u.Unit(entry['unit']).to('nm')}",
        ]

    def _corsika_configuration_iact_parameters(self, parameters_from_db):
        """
        Return CORSIKA IACT parameters.

        Parameters
        ----------
        parameters_from_db : dict
            CORSIKA parameters from the database.

        Returns
        -------
        dict
            Dictionary with CORSIKA IACT parameters.
        """
        parameters = {}
        parameters["SPLIT_AUTO"] = [parameters_from_db["corsika_iact_split_auto"]["value"]]
        parameters["IO_BUFFER"] = [
            self._input_config_io_buff(parameters_from_db["corsika_iact_io_buffer"])
        ]
        parameters["MAX_BUNCHES"] = [parameters_from_db["corsika_iact_max_bunches"]["value"]]
        self._logger.debug(f"IACT parameters: {parameters}")
        return parameters

    def _corsika_configuration_debugging_parameters(self):
        """Return CORSIKA debugging output parameters."""
        return {
            "DEBUG": ["F", 6, "F", 1000000],
            "DATBAS": ["yes"],
            "DIRECT": ["./"],
            "PAROUT": ["F", "F"],
        }

    def _input_config_io_buff(self, entry):
        """Return IO_BUFFER parameter CORSIKA format (Byte or MB required)."""
        value = entry["value"] * u.Unit(entry["unit"]).to("Mbyte")
        # check if value is integer-like
        if value.is_integer():
            return f"{int(value)}MB"
        return f"{int(entry['value'] * u.Unit(entry['unit']).to('byte'))}"

    def _rotate_azimuth_by_180deg(
        self, az, correct_for_geomagnetic_field_alignment=True, invert_operation=False
    ):
        """
        Convert azimuth angle to the CORSIKA coordinate system.

        Corresponds to a rotation by 180 degrees, and optionally a correction for the
        for the differences between the geographic and geomagnetic north pole.

        Parameters
        ----------
        az: float
            Azimuth angle in degrees.
        correct_for_geomagnetic_field_alignment: bool
            Whether to correct for the geomagnetic field alignment.
        invert_operation: bool
            Whether to invert the operation (i.e., convert from CORSIKA to geographic system).

        Returns
        -------
        float
            Azimuth angle in degrees in the CORSIKA coordinate system.
        """
        b_field_declination = 0
        if correct_for_geomagnetic_field_alignment:
            b_field_declination = self.array_model.site_model.get_parameter_value("geomag_rotation")
        if invert_operation:
            return (az - 180 - b_field_declination) % 360
        return (az + 180 + b_field_declination) % 360

    @property
    def primary(self):
        """Primary particle name."""
        return self.primary_particle.name

    def get_config_parameter(self, par_name):
        """
        Get value of CORSIKA configuration parameter.

        Parameters
        ----------
        par_name: str
            Name of the parameter as used in the CORSIKA input file (e.g. PRMPAR, THETAP ...).

        Raises
        ------
        KeyError
            When par_name is not a valid parameter name.

        Returns
        -------
        list
            Value(s) of the parameter.
        """
        par_value = []
        for _, values in self.config.items():
            if par_name in values:
                par_value = values[par_name]
        if len(par_value) == 0:
            raise KeyError(f"Parameter {par_name} is not a CORSIKA config parameter")
        return par_value if len(par_value) > 1 else par_value[0]

    @staticmethod
    def _get_text_single_line(pars, line_begin=""):
        """
        Return one parameter per line for each input parameter.

        Parameters
        ----------
        pars: dict
            Dictionary with the parameters to be written in the file.

        Returns
        -------
        str
            Text with the parameters.
        """
        text = ""
        for par, values in pars.items():
            line = line_begin + par + " "
            for v in values:
                line += str(v) + " "
            line += "\n"
            text += line
        return text

    def generate_corsika_input_file(self, use_multipipe=False, use_test_seeds=False):
        """
        Generate a CORSIKA input file.

        Parameters
        ----------
        use_multipipe: bool
            Whether to set the CORSIKA Inputs file to pipe
            the output directly to sim_telarray.

        """
        if self.is_file_updated:
            self._logger.debug(f"CORSIKA input file already updated: {self.config_file_path}")
            return self.config_file_path
        self._logger.info(f"Exporting CORSIKA input file to {self.config_file_path}")
        _output_generic_file_name = self.set_output_file_and_directory(use_multipipe=use_multipipe)
        self._logger.info(f"Output generic file name: {_output_generic_file_name}")

        with open(self.config_file_path, "w", encoding="utf-8") as file:
            file.write("\n* [ RUN PARAMETERS ]\n")
            text_parameters = self._get_text_single_line(self.config["USER_INPUT"])
            file.write(text_parameters)

            file.write("\n* [ SITE PARAMETERS ]\n")
            text_site_parameters = self._get_text_single_line(
                self.array_model.site_model.get_corsika_site_parameters(
                    config_file_style=True, model_directory=self.array_model.get_config_directory()
                )
            )
            file.write(text_site_parameters)

            file.write("\n* [ IACT ENV PARAMETERS ]\n")
            file.write(f"IACT setenv PRMNAME {self.primary_particle.name}\n")
            file.write(f"IACT setenv ZA {int(self.get_config_parameter('THETAP')[0])}\n")
            file.write(f"IACT setenv AZM {self.azimuth_angle}\n")

            file.write("\n* [ SEEDS ]\n")
            self._write_seeds(file, use_test_seeds)

            file.write("\n* [ TELESCOPES ]\n")
            telescope_list_text = self.get_corsika_telescope_list()
            file.write(telescope_list_text)

            file.write("\n* [ INTERACTION FLAGS ]\n")
            text_interaction_flags = self._get_text_single_line(self.config["INTERACTION_FLAGS"])
            file.write(text_interaction_flags)

            file.write("\n* [ CHERENKOV EMISSION PARAMETERS ]\n")
            text_cherenkov = self._get_text_single_line(
                self.config["CHERENKOV_EMISSION_PARAMETERS"]
            )
            file.write(text_cherenkov)

            file.write("\n* [ DEBUGGING OUTPUT PARAMETERS ]\n")
            text_debugging = self._get_text_single_line(self.config["DEBUGGING_OUTPUT_PARAMETERS"])
            file.write(text_debugging)

            file.write("\n* [ OUTPUT FILE ]\n")
            if use_multipipe:
                run_cta_script = Path(self.config_file_path.parent).joinpath("run_cta_multipipe")
                file.write(f"TELFIL |{run_cta_script!s}\n")
            else:
                file.write(f"TELFIL {_output_generic_file_name}\n")

            file.write("\n* [ IACT TUNING PARAMETERS ]\n")
            text_iact = self._get_text_single_line(
                self.config["IACT_PARAMETERS"],
                "IACT ",
            )
            file.write(text_iact)
            file.write("\nEXIT")

        # Write out the atmospheric transmission file to the model directory.
        # This is done explicitly because it is not done "automatically" when CORSIKA is not piped
        # to sim_telarray.
        self.array_model.site_model.export_atmospheric_transmission_file(
            model_directory=self.array_model.get_config_directory()
        )

        self.is_file_updated = True
        return self.config_file_path

    def get_corsika_config_file_name(self, file_type, run_number=None):
        """
        Get a CORSIKA config style file name for various configuration file types.

        Parameters
        ----------
        file_type: str
            The type of file (determines the file suffix).
            Choices are config_tmp, config or output_generic.
        run_number: int
            Run number.

        Returns
        -------
        str
            for file_type="config_tmp":
                Return CORSIKA input file name for one specific run.
                This is the input file after being pre-processed by sim_telarray (pfp).
            for file_type="config":
                Return generic CORSIKA config input file name.
            for file_type="output_generic"
                Return generic file name for the TELFIL option in the CORSIKA inputs file.
            for file_type="multipipe"
                Return multipipe "file name" for the TELFIL option in the CORSIKA inputs file.

        Raises
        ------
        ValueError
            If file_type is unknown or if the run number is not given for file_type==config_tmp.
        """
        if file_type == "config_tmp" and run_number is None:
            raise ValueError("Must provide a run number for a temporary CORSIKA config file")

        file_label = f"_{self.label}" if self.label is not None else ""

        _vc_low = self.get_config_parameter("VIEWCONE")[0]
        _vc_high = self.get_config_parameter("VIEWCONE")[1]
        view_cone = (
            f"_cone{int(_vc_low):d}-{int(_vc_high):d}" if _vc_low != 0 or _vc_high != 0 else ""
        )

        run_number_in_file_name = ""
        if file_type == "output_generic":
            # The XXXXXX will be replaced by the run number after the pfp step with sed
            run_number_in_file_name = "runXXXXXX_"
        if file_type == "config_tmp":
            run_number_in_file_name = f"run{run_number:06}_"

        base_name = (
            f"{self.primary_particle.name}_{run_number_in_file_name}"
            f"za{int(self.get_config_parameter('THETAP')[0]):02}deg_"
            f"azm{self.azimuth_angle:03}deg{view_cone}_"
            f"{self.array_model.site}_{self.array_model.layout_name}_"
            f"{self.array_model.model_version}{file_label}"
        )

        if file_type == "config_tmp":
            return f"corsika_config_{base_name}.txt"
        if file_type == "config":
            return f"corsika_config_{base_name}.input"
        if file_type == "output_generic":
            return f"{base_name}.corsika.zst"
        if file_type == "multipipe":
            return f"multi_cta-{self.array_model.site}-{self.array_model.layout_name}.cfg"

        raise ValueError(f"The requested file type ({file_type}) is unknown")

    def set_output_file_and_directory(self, use_multipipe=False):
        """
        Set output file names and directories.

        Parameters
        ----------
        use_multipipe: bool
            Whether to set the CORSIKA Inputs file to pipe
            the output directly to sim_telarray. Defines directory names.

        Returns
        -------
        str
            Output file name.
        """
        self.config_file_path = self.io_handler.get_output_file(
            file_name=self.get_corsika_config_file_name(file_type="config"),
            sub_dir="corsika_sim_telarray" if use_multipipe else "corsika",
        )
        return self.get_corsika_config_file_name(file_type="output_generic")

    def _write_seeds(self, file, use_test_seeds=False):
        """
        Generate and write seeds in the CORSIKA input file.

        Parameters
        ----------
        file: stream
            File where the telescope positions will be written.
        """
        random_seed = self.get_config_parameter("PRMPAR") + self.run_number
        rng = np.random.default_rng(random_seed)
        corsika_seeds = [534, 220, 1104, 382]
        if not use_test_seeds:
            corsika_seeds = [int(rng.uniform(0, 1e7)) for _ in range(4)]
        for s in corsika_seeds:
            file.write(f"SEED {s} 0 0\n")

    def get_corsika_telescope_list(self):
        """
        List of telescope positions in the format required for the CORSIKA input file.

        Returns
        -------
        str
            Piece of text to be added to the CORSIKA input file.
        """
        corsika_input_list = ""
        for telescope_name, telescope in self.array_model.telescope_models.items():
            positions = telescope.get_parameter_value_with_unit("array_element_position_ground")
            corsika_input_list += "TELESCOPE"
            for pos in positions:
                corsika_input_list += f"\t {pos.to('cm').value:.3f}"
            sphere_radius = telescope.get_parameter_value_with_unit("telescope_sphere_radius").to(
                "cm"
            )
            corsika_input_list += f"\t {sphere_radius:.3f}"
            corsika_input_list += f"\t # {telescope_name}\n"

        return corsika_input_list

    @property
    def run_number(self):
        """Set run number."""
        return self._run_number

    @run_number.setter
    def run_number(self, run_number):
        """
        Set run number and validate it.

        Parameters
        ----------
        run_number: int
            Run number.
        """
        self._run_number = self.validate_run_number(run_number)

    def validate_run_number(self, run_number):
        """
        Validate run number and return it. Return run number from configuration if None.

        Parameters
        ----------
        run_number: int
            Run number.

        Returns
        -------
        int
            Run number.

        Raises
        ------
        ValueError
            If run_number is not a valid value (< 1 or > 999999).
        """
        if run_number is None:
            return self.run_number
        if not float(run_number).is_integer() or run_number < 1 or run_number > 999999:
            raise ValueError(
                f"Invalid type of run number ({run_number}) - it must be an uint < 1000000."
            )
        return run_number



#!/usr/bin/python3

r"""
    Derives the mirror alignment parameters using cumulative PSF measurement.

    This includes parameters mirror_reflection_random_angle, \
    mirror_align_random_horizontal and mirror_align_random_vertical.

    The measured cumulative PSF should be provided by using the command line argument data. \
    A file name is expected, in which the file should contain 3 columns: radial distance in mm, \
    differential value of photon intensity and its integral value.

    The derivation is performed through gradient descent optimization that minimizes either the \
    Root Mean Squared Deviation (RMSD) between measured and simulated PSF curves (default) or the \
    Kolmogorov-Smirnov (KS) statistic when the --ks_statistic flag is used.

    The optimization workflow includes:

    * Loading and preprocessing PSF data from measurement files
    * Running gradient descent optimization to minimize RMSD
    * Generating cumulative PSF plots for each iteration showing optimization progression
    * Logging parameter evolution through gradient descent steps
    * Creating convergence plots showing RMSD and D80 evolution
    * Automatically generating D80 vs off-axis angle analysis for best parameters
    * Optionally exporting optimized parameters as simulation model files

    The assumption are:

    a) mirror_align_random_horizontal and mirror_align_random_vertical are the same.

    b) mirror_align_random_horizontal/vertical have no dependence on the zenith angle.

    One example of the plot generated by this applications are shown below.

    .. _derive_psf_parameters_plot:
    .. image::  images/gradient_descent.png
      :width: 49 %

    Command line arguments
    ----------------------
    site (str, required)
        North or South.
    telescope (str, required)
        Telescope model name (e.g. LST-1, SST-D, ...).
    model_version (str, optional)
        Model version.
    parameter_version (str, optional)
        Parameter version for model parameter file export.
    src_distance (float, optional)
        Source distance in km.
    zenith (float, optional)
        Zenith angle in deg.
    data (str, optional)
        Name of the data file with the measured cumulative PSF.
    plot_all (activation mode, optional)
        If activated, plots will be generated for all values tested during tuning.
    fixed (activation mode, optional)
        Keep the first entry of mirror_reflection_random_angle fixed.
    test (activation mode, optional)
        If activated, application will be faster by simulating fewer photons.
    write_psf_parameters (activation mode, optional)
        Write the optimized PSF parameters as simulation model parameter files.
    rmsd_threshold (float, optional)
        RMSD threshold for gradient descent convergence (default: 0.007).
    learning_rate (float, optional)
        Learning rate for gradient descent optimization (default: 0.01).
    monte_carlo_analysis (activation mode, optional)
        Run Monte Carlo analysis to find statistical uncertainties.

    Example
    -------
    --telescope LSTN-01 --model_version 6.0.0



    Run the application:

    .. code-block:: console

        simtools-derive-psf-parameters --site North --telescope LSTN-01 \\
            --model_version 6.0.0 --data tests/resources/PSFcurve_data_v2.ecsv --plot_all --test

    Run with parameter export:

    .. code-block:: console

        simtools-derive-psf-parameters --site North --telescope LSTN-01 --model_version 6.0.0 \\
            --plot_all --test --rmsd_threshold 0.01 --learning_rate 0.001 \\
            --data tests/resources/PSFcurve_data_v2.ecsv \\
            --write_psf_parameters

    Run monte carlo analysis:

    .. code-block:: console

        simtools-derive-psf-parameters --site North --telescope LSTN-01 --model_version 6.0.0 \\
            --plot_all --test --monte_carlo_analysis \\
            --data tests/resources/PSFcurve_data_v2.ecsv \\
            --write_psf_parameters

    The output is saved in simtools-output/derive_psf_parameters.

    Output files include:

    * Gradient descent progression log in psf_gradient_descent_[telescope].log
    * Gradient descent convergence plots in gradient_descent_convergence_[telescope].png
    * PSF progression plots showing evolution through iterations (if --plot_all is specified)
    * D80 vs off-axis angle plots (d80_vs_offaxis_cm.png, d80_vs_offaxis_deg.png)
    * Optimized simulation model parameter files (if --write_psf_parameters is specified)

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.model.model_utils import initialize_simulation_models
from simtools.ray_tracing import psf_parameter_optimisation as psf_opt


def _parse():
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Derive mirror_reflection_random_angle, mirror_align_random_horizontal "
            "and mirror_align_random_vertical using cumulative PSF measurement."
        ),
    )
    config.parser.add_argument(
        "--src_distance",
        help="Source distance in km",
        type=float,
        default=10,
    )
    config.parser.add_argument("--zenith", help="Zenith angle in deg", type=float, default=20)
    config.parser.add_argument(
        "--data", help="Data file name with the measured PSF vs radius [cm]", type=str
    )
    config.parser.add_argument(
        "--plot_all",
        help=(
            "On: plot cumulative PSF for all tested combinations, "
            "Off: plot it only for the best set of values"
        ),
        action="store_true",
    )
    config.parser.add_argument(
        "--fixed",
        help=("Keep the first entry of mirror_reflection_random_angle fixed."),
        action="store_true",
    )
    config.parser.add_argument(
        "--write_psf_parameters",
        help=("Write the optimized PSF parameters as simulation model parameter files"),
        action="store_true",
        required=False,
    )
    config.parser.add_argument(
        "--rmsd_threshold",
        help=(
            "RMSD threshold for gradient descent convergence "
            "(not used with --monte_carlo_analysis)."
        ),
        type=float,
        default=0.01,
    )
    config.parser.add_argument(
        "--learning_rate",
        help=(
            "Learning rate for gradient descent optimization "
            "(not used with --monte_carlo_analysis)."
        ),
        type=float,
        default=0.01,
    )
    config.parser.add_argument(
        "--monte_carlo_analysis",
        help="Run analysis to find monte carlo uncertainties.",
        action="store_true",
    )
    config.parser.add_argument(
        "--ks_statistic",
        help="Use KS statistic for monte carlo uncertainty analysis.",
        action="store_true",
    )
    config.parser.add_argument(
        "--fraction",
        help="PSF containment fraction for diameter calculation (e.g., 0.8 for D80, 0.95 for D95).",
        type=float,
        default=0.8,
    )
    config.parser.add_argument(
        "--cleanup",
        help="Remove intermediate *.log and *.lis* files after optimization.",
        action="store_true",
    )
    return config.initialize(
        db_config=True,
        simulation_model=["telescope", "model_version", "parameter_version"],
    )


def main():
    """Derive PSF parameters."""
    app_context = startup_application(_parse)

    tel_model, site_model, _ = initialize_simulation_models(
        label=app_context.args.get("label"),
        db_config=app_context.db_config,
        site=app_context.args["site"],
        telescope_name=app_context.args["telescope"],
        model_version=app_context.args["model_version"],
    )

    psf_opt.run_psf_optimization_workflow(
        tel_model,
        site_model,
        app_context.args,
        app_context.io_handler.get_output_directory(),
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3

r"""
    Generates a set of histograms Cherenkov photon distributions from CORSIKA output.

    The Cherenkov photons (from observation level) are read from a CORSIKA IACT
    output file provided as input.

    The histograms can be saved both into pdfs and in a hdf5 file.

    The following 2D histograms are generated:
        - Number of Cherenkov photons on the ground;
        - Density of Cherenkov photons on the ground;
        - Incoming direction (directive cosines) of the Cherenkov photons;
        - Time of arrival (ns) vs altitude of production (km);
        - Number of Cherenkov photons per event per telescope.

    The following 1D histograms are generated:
        - Wavelength;
        - Counts;
        - Density;
        - Time of arrival;
        - Altitude of production;
        - Number of photons per telescope;
        - Number of photons per event.

    Histograms for the distribution of CORSIKA event header elements can also be generated by using
    the --event_1d_histograms and --event_2d_histograms arguments. The accepted arguments (keys)
    are to be found in the CORSIKA manual, e.g., "total_energy", "zenith", "azimuth".

    Command line arguments
    ----------------------
    iact_file (str, required)
        The name of the CORSIKA IACT file resulted from the CORSIKA simulation.

    telescope_indices (list, optional)
        The list with the telescope indices to be considered in the generation of the histograms.
        Telescopes that are not in the list will not contribute with photons to the histograms.
        If the argument is not given, all telescopes are considered.

    individual_telescopes (bool, optional)
        Indicates whether single histograms are generated for the individual telescopes, or if
        a master histogram is generated for all the telescopes together.
        If the argument is not given, the Cherenkov photons from the given telescopes are considered
        together in the same histograms.

    hist_config (hdf5 or dict, optional)
        The configuration used for generating the histograms.
        It includes information about the bin sizes, the ranges, scale of the plot and units.
        By construction, three major histograms are created to start with:
        - hist_direction (2D): Directive cosines (x and y) for the incoming photons;
        - hist_position (3D): position x, position y, and wavelength;
        - hist_time_altitude (2D): time of arrival and altitude of emission;

    If the argument is not given, the default configuration is generated:

    .. code-block:: console

        hist_direction:
            x axis: {bins: 100, scale: linear, start: -1, stop: 1}
            y axis: {bins: 100, scale: linear, start: -1, stop: 1}

        hist_position:
            x axis:
                bins: 100
                scale: linear
                start: !astropy.units.Quantity
                    unit: &id001 m
                    value: -1000.0
                stop: &id002 !astropy.units.Quantity
                    unit: *id001
                    value: 1000.0
            y axis:
                bins: 100
                scale: linear
                start: !astropy.units.Quantity
                    unit: *id001
                    value: -1000.0
                stop: *id002
            z axis:
                bins: 80
                scale: linear
                start: !astropy.units.Quantity
                    unit: nm
                    value: 200.0
                stop: !astropy.units.Quantity
                    unit: *id003
                    value: 1000.0
        hist_time_altitude:
            x axis:
                bins: 100
                scale: linear
                start: !astropy.units.Quantity
                    unit: ns
                    value: -2000.0
                stop: !astropy.units.Quantity
                    unit: *id004
                    value: 2000.0
            y axis:
                bins: 100
                scale: linear
                start: !astropy.units.Quantity
                    unit: km
                    value: 120.0
                stop: !astropy.units.Quantity
                    unit: *id005
                    value: 0.0


    pdf (bool, optional)
        If set, histograms are saved into pdf files.
        One pdf file contains all the histograms for the Cherenkov photons.
        The name of the file is controlled via hdf5_file_name.
        If event_1d_histograms and event_2d_histograms are used, two separate pdf files might be
        created to accommodate the histograms for the CORSIKA event header elements. The core names
        of these output pdf files are also given by hdf5_file_name argument with the addition of
        'event_1d_histograms' and 'event_2d_histograms'.


    hdf5 (bool, optional)
        If set, histograms are saved into hdf5 files.

    hdf5_file_name (str, optional)
        The name of the output hdf5 data (without the path).
        It requires the --hdf5 flag.
        If not given, hdf5_file_name takes the name from the input IACT file (input_file).
        If the output hdf5_file_name file already exists, the tables associated to the chosen
        flags (e.g. hdf5, event_1d_histograms, event_2d_histograms) will be overwritten. The
        remaining tables, if any, will stay untouched.


    event_1d_histograms (str, optional)
        Generate 1D histograms for elements given in --event_1d_histograms from the CORSIKA event
        header and save into hdf5/pdf files.
        It allows more than one argument, separated by simple spaces.
        Usage: --event_1d_histograms first_interaction_height total_energy.

    event_2d_histograms (str, optional)
        Generate 2D histograms for elements given in --event_2d_histograms from the CORSIKA event
        header and save into hdf5/pdf files.
        It allows more than one argument, separated by simple spaces.
        The elements are grouped into pairs and the 2D histograms are generated always for two
        subsequent elements.
        For example, --event_2d_histograms first_interaction_height total_energy zenith azimuth
        will generate one 2D histogram for first_interaction_height total_energy and another 2D
        histogram for zenith and azimuth.

    Example
    -------
    Generate the histograms for a test IACT file:

     .. code-block:: console

        simtools-generate-corsika-histograms --iact_file /workdir/external/simtools/\\
        tests/resources/tel_output_10GeV-2-gamma-20deg-CTAO-South.corsikaio \\
            --pdf --hdf5 \\
            --event_2d_histograms zenith azimuth --event_1d_histograms total_energy

    Expected final print-out message:

    .. code-block:: console

        INFO::generate_corsika_histograms(l358)::main::Finalizing the application.
        Total time needed: 8s.
"""

import time

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.corsika.corsika_histograms import CorsikaHistograms


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Generate histograms for the Cherenkov photons saved in the CORSIKA IACT file.",
    )

    config.parser.add_argument(
        "--iact_file",
        help="Name of the CORSIKA IACT file from which to generate the histograms.",
        type=str,
        required=True,
    )

    config.parser.add_argument(
        "--telescope_indices",
        help="List of telescope indices to be considered in the generation of the histograms",
        type=str,
        required=False,
        nargs="+",
        default=None,
    )

    config.parser.add_argument(
        "--individual_telescopes",
        help="if False, the histograms are filled for all given telescopes together, otherwise"
        "one histogram is set for each telescope separately.",
        action="store_true",
        required=False,
        default=False,
    )

    config.parser.add_argument(
        "--hist_config",
        help="hdf5 file with the configuration parameters to create the histograms.",
        type=str,
        required=False,
        default=None,
    )

    config.parser.add_argument(
        "--pdf", help="Save histograms into a pdf file.", action="store_true", required=False
    )

    config.parser.add_argument(
        "--hdf5", help="Save histograms into hdf5 files.", action="store_true", required=False
    )

    config.parser.add_argument(
        "--hdf5_file_name",
        help="Name of the hdf5 file where to save the histograms.",
        type=str,
        required=False,
        default=None,
    )

    config.parser.add_argument(
        "--event_1d_histograms",
        help="The keys from the CORSIKA event header to be used for the generation of 1D "
        "histograms. The available choices can been found in the all_event_keys attribute of"
        "the CorsikaHistograms.",
        required=False,
        default=None,
        nargs="*",
    )

    config.parser.add_argument(
        "--event_2d_histograms",
        help="The keys from the CORSIKA event header to be used for the generation of 2D "
        "histograms. The available choices can been found in the all_event_keys attribute of"
        "the CorsikaHistograms.",
        required=False,
        default=None,
        nargs="*",
    )

    config_parser, _ = config.initialize(db_config=False, paths=True)

    if (
        not config_parser["pdf"]
        and not config_parser["hdf5"]
        and not config_parser["event_1d_histograms"]
        and not config_parser["event_2d_histograms"]
    ):
        config.parser.error(
            "At least one argument is required: --pdf, --hdf5, --event_1d_histograms, or "
            "--event_2d_histograms."
        )

    return config_parser, _


def main():
    """Generate a set of histograms for the Cherenkov photons saved in the CORSIKA IACT file."""
    app_context = startup_application(_parse)

    initial_time = time.time()

    corsika_histograms_instance = CorsikaHistograms(
        app_context.args["iact_file"],
        output_path=app_context.io_handler.get_output_directory(),
        hdf5_file_name=app_context.args["hdf5_file_name"],
    )
    corsika_histograms_instance.run_export_pipeline(
        individual_telescopes=app_context.args["individual_telescopes"],
        hist_config=app_context.args["hist_config"],
        indices_arg=app_context.args["telescope_indices"],
        write_pdf=app_context.args["pdf"],
        write_hdf5=app_context.args["hdf5"],
        event1d=app_context.args["event_1d_histograms"],
        event2d=app_context.args["event_2d_histograms"],
        test=app_context.args["test"],
    )

    final_time = time.time()
    app_context.logger.info(
        f"Finalizing the application. Total time needed: {round(final_time - initial_time)}s."
    )


if __name__ == "__main__":
    main()



"""Light emission simulation (e.g. illuminators or flashers)."""

import logging
import shutil
import stat
import subprocess
from pathlib import Path

import astropy.units as u
import numpy as np

from simtools.io import io_handler
from simtools.model.model_utils import initialize_simulation_models
from simtools.runners.simtel_runner import SimtelRunner
from simtools.simtel.simtel_config_writer import SimtelConfigWriter
from simtools.utils.general import clear_default_sim_telarray_cfg_directories
from simtools.utils.geometry import fiducial_radius_from_shape


class SimulatorLightEmission(SimtelRunner):
    """
    Light emission simulation (e.g. illuminators or flashers).

    Uses the sim_telarray LightEmission package to simulate the light emission.

    Parameters
    ----------
    light_emission_config : dict, optional
        Configuration for the light emission (e.g. number of events, model names)
    label : str, optional
        Label for the simulation
    """

    def __init__(self, light_emission_config, db_config=None, label=None):
        """Initialize SimulatorLightEmission."""
        self._logger = logging.getLogger(__name__)
        self.io_handler = io_handler.IOHandler()

        super().__init__(
            simtel_path=light_emission_config.get("simtel_path"), label=label, corsika_config=None
        )

        self.output_directory = self.io_handler.get_output_directory()

        self.telescope_model, self.site_model, self.calibration_model = (
            initialize_simulation_models(
                label=label,
                db_config=db_config,
                site=light_emission_config.get("site"),
                telescope_name=light_emission_config.get("telescope"),
                calibration_device_name=light_emission_config.get("light_source"),
                model_version=light_emission_config.get("model_version"),
            )
        )
        self.telescope_model.write_sim_telarray_config_file(additional_models=self.site_model)

        self.light_emission_config = self._initialize_light_emission_configuration(
            light_emission_config
        )

    def _initialize_light_emission_configuration(self, config):
        """Initialize light emission configuration."""
        if self.calibration_model.get_parameter_value("flasher_type"):
            config["light_source_type"] = self.calibration_model.get_parameter_value(
                "flasher_type"
            ).lower()

        config["flasher_photons"] = (
            self.calibration_model.get_parameter_value("flasher_photons")
            if not config.get("test", False)
            else 1e8
        )

        if config.get("light_source_position") is not None:
            config["light_source_position"] = (
                np.array(config["light_source_position"], dtype=float) * u.m
            )

        return config

    def simulate(self):
        """
        Simulate light emission.

        Returns
        -------
        Path
            The output simtel file path.
        """
        run_script = self.prepare_script()
        log_path = Path(self.output_directory) / "logfile.log"
        with open(log_path, "w", encoding="utf-8") as fh:
            subprocess.run(
                run_script,
                shell=False,
                check=False,
                text=True,
                stdout=fh,
                stderr=fh,
            )
        out = Path(self._get_simulation_output_filename())
        if not out.exists():
            self._logger.warning(f"Expected sim_telarray output not found: {out}")
        return out

    def prepare_script(self):
        """
        Build and return bash run script containing the light-emission command.

        Returns
        -------
        Path
            Full path of the run script.
        """
        script_dir = self.output_directory.joinpath("scripts")
        script_dir.mkdir(parents=True, exist_ok=True)

        app_name = self._get_light_emission_application_name()
        script_file = script_dir / f"{app_name}-light_emission.sh"
        self._logger.debug(f"Run bash script - {script_file}")

        target_out = Path(self._get_simulation_output_filename())
        if target_out.exists():
            raise FileExistsError(
                f"sim_telarray output file exists, cancelling simulation: {target_out}"
            )

        lines = [
            "#!/usr/bin/env bash\n",
            f"{self._make_light_emission_script()}\n\n",
            (
                f"[ -s '{self.output_directory}/{app_name}.iact.gz' ] || "
                f"{{ echo 'LightEmission did not produce IACT file' >&2; exit 1; }}\n\n"
            ),
            f"{self._make_simtel_script()}\n\n",
            f"rm -f '{self.output_directory}/{app_name}.iact.gz'\n\n",
        ]

        script_file.write_text("".join(lines), encoding="utf-8")
        script_file.chmod(script_file.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP)
        return script_file

    def _get_prefix(self):
        prefix = self.light_emission_config.get("output_prefix", "")
        if prefix is not None:
            return f"{prefix}_"
        return ""

    def _get_light_emission_application_name(self):
        """
        Return the LightEmission application and mode from type.

        Returns
        -------
        str
            app_name
        """
        if self.light_emission_config["light_source_type"] == "flat_fielding":
            return "ff-1m"
        # default to illuminator xyzls, mode from setup
        return "xyzls"

    def _get_telescope_pointing(self):
        """
        Return telescope pointing based on light source type.

        For flat_fielding sims, avoid calibration pointing entirely; default angles to (0,0).

        Returns
        -------
        tuple
            The telescope pointing angles (theta, phi).

        """
        if self.light_emission_config["light_source_type"] == "flat_fielding":
            return 0.0, 0.0
        if self.light_emission_config.get("light_source_position") is not None:
            self._logger.info("Using fixed (vertical up) telescope pointing.")
            return 0.0, 0.0
        _, angles = self._calibration_pointing_direction()
        return angles[0], angles[1]

    def _calibration_pointing_direction(self, x_cal=None, y_cal=None, z_cal=None):
        """
        Calculate the pointing of the calibration device towards the telescope.

        This is for calibration devices not installed on telescopes (e.g. illuminators).

        Returns
        -------
        list
            The pointing vector from the calibration device to the telescope.
        """
        if x_cal is None or y_cal is None or z_cal is None:
            x_cal, y_cal, z_cal = self.calibration_model.get_parameter_value_with_unit(
                "array_element_position_ground"
            )
        x_cal, y_cal, z_cal = [coord.to(u.m).value for coord in (x_cal, y_cal, z_cal)]
        cal_vect = np.array([x_cal, y_cal, z_cal])
        x_tel, y_tel, z_tel = self.telescope_model.get_parameter_value_with_unit(
            "array_element_position_ground"
        )
        x_tel, y_tel, z_tel = [coord.to(u.m).value for coord in (x_tel, y_tel, z_tel)]
        tel_vect = np.array([x_tel, y_tel, z_tel])

        direction_vector = tel_vect - cal_vect
        # pointing vector from calibration device to telescope
        pointing_vector = np.round(direction_vector / np.linalg.norm(direction_vector), 6)

        # Calculate telescope theta and phi angles
        tel_theta = 180 - np.round(
            np.rad2deg(np.arccos(direction_vector[2] / np.linalg.norm(direction_vector))), 6
        )
        tel_phi = 180 - np.round(
            np.rad2deg(np.arctan2(direction_vector[1], direction_vector[0])), 6
        )
        # Calculate source beam theta and phi angles
        direction_vector_inv = direction_vector * -1
        source_theta = np.round(
            np.rad2deg(np.arccos(direction_vector_inv[2] / np.linalg.norm(direction_vector_inv))),
            6,
        )
        source_phi = np.round(
            np.rad2deg(np.arctan2(direction_vector_inv[1], direction_vector_inv[0])), 6
        )
        return pointing_vector.tolist(), [tel_theta, tel_phi, source_theta, source_phi]

    def _write_telescope_position_file(self):
        """
        Write the telescope positions to a telescope_position file.

        The file will contain lines in the format: x y z r in cm

        Returns
        -------
        Path
            The path to the generated telescope_position file.
        """
        x_tel, y_tel, z_tel = self.telescope_model.get_parameter_value_with_unit(
            "array_element_position_ground"
        )
        x_tel, y_tel, z_tel = [coord.to(u.cm).value for coord in (x_tel, y_tel, z_tel)]

        radius = self.telescope_model.get_parameter_value_with_unit("telescope_sphere_radius")
        radius = radius.to(u.cm).value  # Convert radius to cm

        telescope_position_file = self.output_directory.joinpath("telescope_position.dat")
        telescope_position_file.write_text(f"{x_tel} {y_tel} {z_tel} {radius}\n", encoding="utf-8")
        return telescope_position_file

    def _prepare_flasher_atmosphere_files(self, config_directory, model_id=1):
        """
        Prepare canonical atmosphere aliases for ff-1m and return model id.

        The ff-1m tool requires atmosphere files atmprof1.dat or atm_profile_model_1.dat and
        as configuration parameter the atmosphere id ('--atmosphere id').

        """
        src_path = config_directory / self.site_model.get_parameter_value("atmospheric_profile")
        self._logger.debug(f"Using atmosphere profile: {src_path}")

        for name in (f"atmprof{model_id}.dat", f"atm_profile_model_{model_id}.dat"):
            dst = config_directory / name
            try:
                if dst.exists() or dst.is_symlink():
                    dst.unlink()
                try:
                    dst.symlink_to(src_path)
                except OSError:
                    shutil.copy2(src_path, dst)
            except OSError as copy_err:
                self._logger.warning(f"Failed to create atmosphere alias {dst.name}: {copy_err}")
        return model_id

    def _make_light_emission_script(self):
        """
        Create the light emission script to run the light emission package.

        Require the specified pre-compiled light emission package application
        in the sim_telarray/LightEmission/ path.

        Returns
        -------
        str
            The commands to run the Light Emission package
        """
        config_directory = self.io_handler.get_model_configuration_directory(
            model_version=self.site_model.model_version
        )
        app_name = self._get_light_emission_application_name()
        corsika_observation_level = self.site_model.get_parameter_value_with_unit(
            "corsika_observation_level"
        )

        parts = [str(self._simtel_path / "sim_telarray/LightEmission") + f"/{app_name}"]
        parts.extend(self._get_site_command(app_name, config_directory, corsika_observation_level))
        parts.extend(self._get_light_source_command())
        if self.light_emission_config["light_source_type"] == "illuminator":
            parts += [
                "-A",
                (
                    f"{config_directory}/"
                    f"{self.telescope_model.get_parameter_value('atmospheric_profile')}"
                ),
            ]
        parts += [f"-o {self.output_directory}/{app_name}.iact.gz", "\n"]
        return " ".join(parts)

    def _get_site_command(self, app_name, config_directory, corsika_observation_level):
        """Return site command with altitude, atmosphere and telescope_position handling."""
        if app_name in ("ff-1m",):
            atmo_id = self._prepare_flasher_atmosphere_files(config_directory)
            return [
                "-I.",
                f"-I{self._simtel_path / 'sim_telarray/cfg'}",
                f"-I{config_directory}",
                f"--altitude {corsika_observation_level.to(u.m).value}",
                f"--atmosphere {atmo_id}",
            ]
        # default path (not used for flasher now, but kept for completeness)
        return [
            f"-h  {corsika_observation_level.to(u.m).value} ",
            f"--telpos-file {self._write_telescope_position_file()}",
        ]

    def _get_light_source_command(self):
        """Return light-source specific command options."""
        if self.light_emission_config["light_source_type"] == "flat_fielding":
            return self._add_flasher_command_options()
        if self.light_emission_config["light_source_type"] == "illuminator":
            return self._add_illuminator_command_options()
        raise ValueError(
            f"Unknown light_source_type '{self.light_emission_config['light_source_type']}'"
        )

    def _add_flasher_command_options(self):
        """Add flasher options for all telescope types (ff-1m style)."""
        flasher_xyz = self.calibration_model.get_parameter_value_with_unit("flasher_position")
        camera_diam_cm = (
            self.telescope_model.get_parameter_value_with_unit("camera_body_diameter")
            .to(u.cm)
            .value
        )
        camera_shape = self.telescope_model.get_parameter_value("camera_body_shape")
        camera_radius = fiducial_radius_from_shape(camera_diam_cm, camera_shape)
        flasher_wavelength = self.calibration_model.get_parameter_value_with_unit(
            "flasher_wavelength"
        )
        dist_cm = self.calculate_distance_focal_plane_calibration_device().to(u.cm).value
        angular_distribution = self._get_angular_distribution_string_for_sim_telarray()

        # Build pulse table for ff-1m using model width/exp parameters; else use token.
        pulse_arg = self._get_pulse_shape_string_for_sim_telarray()
        pulse_shape = self.calibration_model.get_parameter_value("flasher_pulse_shape")
        width_q = self.calibration_model.get_parameter_value_with_unit("flasher_pulse_width")
        exp_q = self.calibration_model.get_parameter_value_with_unit("flasher_pulse_exp_decay")
        if (
            isinstance(exp_q, u.Quantity)
            and isinstance(width_q, u.Quantity)
            and pulse_shape == "Gauss-Exponential"
        ):
            try:
                base_dir = self.io_handler.get_output_directory("pulse_shapes")

                def _sanitize_name(value):
                    return "".join(
                        ch if (ch.isalnum() or ch in ("-", "_")) else "_" for ch in str(value)
                    )

                tel = self.light_emission_config.get("telescope") or "telescope"
                cal = self.light_emission_config.get("light_source") or "calibration"
                fname = f"flasher_pulse_shape_{_sanitize_name(tel)}_{_sanitize_name(cal)}.dat"
                table_path = base_dir / fname
                fadc_bins = self.telescope_model.get_parameter_value("fadc_sum_bins")

                SimtelConfigWriter.write_lightpulse_table_gauss_expconv(
                    file_path=table_path,
                    width_ns=width_q.to(u.ns).value,
                    exp_decay_ns=exp_q.to(u.ns).value,
                    fadc_sum_bins=fadc_bins,
                    time_margin_ns=5.0,
                )
                pulse_arg = str(table_path)
            except (ValueError, OSError) as err:
                self._logger.warning(f"Failed to write pulse shape table, using token: {err}")

        return [
            f"--events {self.light_emission_config['number_of_events']}",
            f"--photons {self.light_emission_config['flasher_photons']}",
            f"--bunchsize {self.calibration_model.get_parameter_value('flasher_bunch_size')}",
            f"--xy {flasher_xyz[0].to(u.cm).value},{flasher_xyz[1].to(u.cm).value}",
            f"--distance {dist_cm}",
            f"--camera-radius {camera_radius}",
            f"--spectrum {int(flasher_wavelength.to(u.nm).value)}",
            f"--lightpulse {pulse_arg}",
            f"--angular-distribution {angular_distribution}",
        ]

    def _add_illuminator_command_options(self):
        """Get illuminator-specific command options for light emission script."""
        pos = self.light_emission_config.get("light_source_position")
        if pos is None:
            pos = self.calibration_model.get_parameter_value_with_unit(
                "array_element_position_ground"
            )
        x_cal, y_cal, z_cal = pos
        if self.light_emission_config.get("light_source_pointing"):
            pointing_vector = self.light_emission_config["light_source_pointing"]
        else:
            pointing_vector = self._calibration_pointing_direction(x_cal, y_cal, z_cal)[0]
        flasher_wavelength = self.calibration_model.get_parameter_value_with_unit(
            "flasher_wavelength"
        )
        angular_distribution = self._get_angular_distribution_string_for_sim_telarray()

        return [
            f"-x {x_cal.to(u.cm).value}",
            f"-y {y_cal.to(u.cm).value}",
            f"-z {z_cal.to(u.cm).value}",
            f"-d {','.join(map(str, pointing_vector))}",
            f"-n {self.light_emission_config['flasher_photons']}",
            f"-s {int(flasher_wavelength.to(u.nm).value)}",
            f"-p {self._get_pulse_shape_string_for_sim_telarray()}",
            f"-a {angular_distribution}",
        ]

    def _make_simtel_script(self):
        """
        Return the command to run sim_telarray using the output from the previous step.

        Returns
        -------
        str
            The command to run sim_telarray
        """
        theta, phi = self._get_telescope_pointing()

        simtel_bin = self._simtel_path.joinpath("sim_telarray/bin/sim_telarray/")

        parts = [
            f"{simtel_bin}",
            f"-I{self.telescope_model.config_file_directory}",
            f"-I{simtel_bin}",
            f"-c {self.telescope_model.config_file_path}",
            "-DNUM_TELESCOPES=1",
            super().get_config_option(
                "altitude",
                self.site_model.get_parameter_value_with_unit("corsika_observation_level")
                .to(u.m)
                .value,
            ),
            super().get_config_option(
                "atmospheric_transmission",
                self.site_model.get_parameter_value("atmospheric_transmission"),
            ),
            super().get_config_option("TRIGGER_TELESCOPES", "1"),
            super().get_config_option("TELTRIG_MIN_SIGSUM", "2"),
            super().get_config_option("PULSE_ANALYSIS", "-30"),
            super().get_config_option("MAXIMUM_TELESCOPES", 1),
            super().get_config_option("telescope_theta", f"{theta}"),
            super().get_config_option("telescope_phi", f"{phi}"),
        ]

        if self.light_emission_config["light_source_type"] == "flat_fielding":
            parts.append(super().get_config_option("Bypass_Optics", "1"))

        app_name = self._get_light_emission_application_name()
        pref = self._get_prefix()
        parts += [
            super().get_config_option("power_law", "2.68"),
            super().get_config_option("input_file", f"{self.output_directory}/{app_name}.iact.gz"),
            super().get_config_option(
                "output_file", f"{self.output_directory}/{pref}{app_name}.simtel.zst"
            ),
            super().get_config_option(
                "histogram_file", f"{self.output_directory}/{pref}{app_name}.ctsim.hdata\n"
            ),
        ]

        return clear_default_sim_telarray_cfg_directories(" ".join(parts))

    def _get_simulation_output_filename(self):
        """Get the filename of the simulation output."""
        app_name = self._get_light_emission_application_name()
        pref = self._get_prefix()
        return f"{self.output_directory}/{pref}{app_name}.simtel.zst"

    def calculate_distance_focal_plane_calibration_device(self):
        """
        Calculate distance between focal plane and calibration device.

        For flasher-type light sources. Flasher position is given in mirror coordinates,
        with positive z pointing towards the camera, so the distance is focal_length - flasher_z.

        Returns
        -------
        astropy.units.Quantity
            Distance between calibration device and focal plane.
        """
        focal_length = self.telescope_model.get_parameter_value_with_unit("focal_length").to(u.m)
        flasher_z = self.calibration_model.get_parameter_value_with_unit("flasher_position")[2].to(
            u.m
        )
        return focal_length - flasher_z

    def _get_angular_distribution_string_for_sim_telarray(self):
        """
        Get the angular distribution string for sim_telarray.

        Returns
        -------
        str
            The angular distribution string.
        """
        opt = self.calibration_model.get_parameter_value("flasher_angular_distribution")
        option_string = str(opt).lower() if opt is not None else ""
        width = self.calibration_model.get_parameter_value_with_unit(
            "flasher_angular_distribution_width"
        )
        return f"{option_string}:{width.to(u.deg).value}" if width is not None else option_string

    def _get_pulse_shape_string_for_sim_telarray(self):
        """
        Get the pulse shape string for sim_telarray.

        Returns
        -------
        str
            The pulse shape string.
        """
        opt = self.calibration_model.get_parameter_value("flasher_pulse_shape")
        option_string = str(opt).lower() if opt is not None else ""
        width = self.calibration_model.get_parameter_value_with_unit("flasher_pulse_width")
        return f"{option_string}:{width.to(u.ns).value}" if width is not None else option_string



#!/usr/bin/python3
"""Plot array elements for a layout."""

from collections import Counter
from typing import NamedTuple

import astropy.units as u
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from astropy.table import Column
from matplotlib.collections import PatchCollection

from simtools.utils import geometry as transf
from simtools.utils import names
from simtools.visualization import legend_handlers as leg_h


class PlotBounds(NamedTuple):
    """Axis-aligned bounds for the layout in meters.

    Attributes
    ----------
    x_lim : tuple[float, float]
        Min/max for x (meters).
    y_lim : tuple[float, float]
        Min/max for y (meters).
    """

    x_lim: tuple[float, float]
    y_lim: tuple[float, float]


def plot_array_layout(
    telescopes,
    show_tel_label=False,
    axes_range=None,
    marker_scaling=1.0,
    background_telescopes=None,
    grayed_out_elements=None,
    highlighted_elements=None,
    legend_location="best",
    bounds_mode="exact",
    padding=0.1,
    x_lim=None,
    y_lim=None,
):
    """
    Plot telescope array layout.

    Parameters
    ----------
    telescopes : Table
        Telescope data table.
    show_tel_label : bool
        Show telescope labels (default False).
    axes_range : float or None
        Axis range, auto if None.
    marker_scaling : float
        Marker size scale factor.
    background_telescopes : Table or None
        Optional background telescope table.
    grayed_out_elements : list or None
        List of telescope names to plot as gray circles.
    highlighted_elements : list or None
        List of telescope names to plot with red circles around them.
    legend_location : str
        Location of the legend (default "best").

    Returns
    -------
    fig : Figure
        Matplotlib figure object.

    Other Parameters
    ----------------
    bounds_mode : {"symmetric", "exact"}
        Controls axis limits calculation. "symmetric" uses +-R where R is the padded
        maximum extent (default), while "exact" uses individual x/y min/max bounds.
    padding : float
        Fractional padding applied around computed extents (used for both modes).
    x_lim, y_lim : tuple(float, float), optional
        Explicit axis limits in meters. If provided, these override axes_range and bounds_mode
        for the respective axis. If only one is provided, the other axis is derived per mode.
    """
    fig, ax = plt.subplots(1)

    # If explicit limits are provided (one or both), filter patches accordingly
    filter_x = x_lim
    filter_y = y_lim

    patches, plot_range, highlighted_patches, bounds = get_patches(
        ax,
        telescopes,
        show_tel_label,
        axes_range,
        marker_scaling,
        grayed_out_elements,
        highlighted_elements,
        filter_x_lim=filter_x,
        filter_y_lim=filter_y,
    )

    plot_range, bounds = _get_patches_for_background_telescopes(
        ax,
        background_telescopes,
        axes_range,
        marker_scaling,
        bounds_mode,
        plot_range,
        bounds,
        filter_x_lim=filter_x,
        filter_y_lim=filter_y,
    )

    if legend_location != "no_legend":
        update_legend(ax, telescopes, grayed_out_elements, legend_location)

    x_lim, y_lim = _get_axis_limits(
        axes_range, bounds_mode, padding, plot_range, bounds, x_lim, y_lim
    )

    finalize_plot(ax, patches, "Easting [m]", "Northing [m]", x_lim, y_lim, highlighted_patches)

    return fig


def _get_axis_limits(
    axes_range,
    bounds_mode,
    padding,
    plot_range,
    bounds,
    x_lim_override=None,
    y_lim_override=None,
):
    """Get axis limits based on mode and padding."""

    def _derive_axis(axis: str) -> tuple[float, float]:
        if bounds_mode == "exact":
            if axis == "x":
                span = bounds.x_lim[1] - bounds.x_lim[0]
                pad = padding * span
                return (bounds.x_lim[0] - pad, bounds.x_lim[1] + pad)
            span = bounds.y_lim[1] - bounds.y_lim[0]
            pad = padding * span
            return (bounds.y_lim[0] - pad, bounds.y_lim[1] + pad)
        # symmetric
        sym = plot_range
        padf = max(0.0, min(1.0, float(padding))) if padding is not None else 0.0
        sym *= 1.0 + padf
        return (-sym, sym)

    # Highest priority: explicit overrides (per axis)
    if x_lim_override is not None or y_lim_override is not None:
        x_lim = x_lim_override if x_lim_override is not None else _derive_axis("x")
        y_lim = y_lim_override if y_lim_override is not None else _derive_axis("y")
        return x_lim, y_lim

    if axes_range is not None:
        return (-axes_range, axes_range), (-axes_range, axes_range)
    # Derive both axes using selected mode
    return _derive_axis("x"), _derive_axis("y")


def _get_patches_for_background_telescopes(
    ax,
    background_telescopes,
    axes_range,
    marker_scaling,
    bounds_mode,
    plot_range,
    bounds,
    filter_x_lim=None,
    filter_y_lim=None,
):
    """Get background telescope patches and update plot range/bounds."""
    if background_telescopes is None:
        return plot_range, bounds

    bg_patches, bg_range, _, bg_bounds = get_patches(
        ax,
        background_telescopes,
        False,
        axes_range,
        marker_scaling,
        None,
        None,
        filter_x_lim=filter_x_lim,
        filter_y_lim=filter_y_lim,
    )
    ax.add_collection(PatchCollection(bg_patches, match_original=True, alpha=0.1))
    if axes_range is None:
        if bounds_mode == "symmetric":
            plot_range = max(plot_range, bg_range)
        else:
            bounds = PlotBounds(
                x_lim=(
                    min(bounds.x_lim[0], bg_bounds.x_lim[0]),
                    max(bounds.x_lim[1], bg_bounds.x_lim[1]),
                ),
                y_lim=(
                    min(bounds.y_lim[0], bg_bounds.y_lim[0]),
                    max(bounds.y_lim[1], bg_bounds.y_lim[1]),
                ),
            )
    return plot_range, bounds


def _apply_limits_filter(telescopes, pos_x, pos_y, filter_x_lim, filter_y_lim):
    """Filter telescope table and positions by optional axis limits."""
    if filter_x_lim is None and filter_y_lim is None:
        return telescopes, pos_x, pos_y

    px = np.asarray(pos_x.to_value(u.m))
    py = np.asarray(pos_y.to_value(u.m))
    mask = np.ones(px.shape, dtype=bool)
    if filter_x_lim is not None:
        mask &= (px >= float(filter_x_lim[0])) & (px <= float(filter_x_lim[1]))
    if filter_y_lim is not None:
        mask &= (py >= float(filter_y_lim[0])) & (py <= float(filter_y_lim[1]))

    if mask.size and mask.any():
        return telescopes[mask], pos_x[mask], pos_y[mask]

    # No telescopes within limits
    return telescopes[:0], pos_x[:0], pos_y[:0]


def get_patches(
    ax,
    telescopes,
    show_tel_label,
    axes_range,
    marker_scaling,
    grayed_out_elements=None,
    highlighted_elements=None,
    filter_x_lim=None,
    filter_y_lim=None,
):
    """
    Get plot patches and axis range.

    Parameters
    ----------
    ax : Axes
        Matplotlib axes object.
    telescopes : Table
        Telescope data table.
    show_tel_label : bool
        Show telescope labels.
    axes_range : float or None
        Axis range, auto if None.
    marker_scaling : float
        Marker size scale factor.
    grayed_out_elements : list or None
        List of telescope names to plot as gray circles.
    highlighted_elements : list or None
        List of telescope names to plot with red circles around them.

    Returns
    -------
    patches : list
        List of telescope patches.
    axes_range : float
        Calculated or input symmetric axis range (meters).
    highlighted_patches : list
        List of highlighted telescope patches.
    bounds : PlotBounds
        Min/max for x and y in meters.
    """
    pos_x, pos_y = get_positions(telescopes)
    tel_table, pos_x, pos_y = _apply_limits_filter(
        telescopes, pos_x, pos_y, filter_x_lim, filter_y_lim
    )

    tel_table["pos_x_rotated"] = Column(pos_x)
    tel_table["pos_y_rotated"] = Column(pos_y)

    patches, radii, highlighted_patches = create_patches(
        tel_table, marker_scaling, show_tel_label, ax, grayed_out_elements, highlighted_elements
    )

    if len(radii) == 0:
        r = 0.0
    else:
        radii_q = u.Quantity(radii)
        r = float(np.nanmax(radii_q).to_value(u.m))

    if len(pos_x) == 0:
        bounds = PlotBounds(x_lim=(0.0, 0.0), y_lim=(0.0, 0.0))
        if axes_range:
            return patches, axes_range, highlighted_patches, bounds
        return patches, 0.0, highlighted_patches, bounds

    x_min = float(np.nanmin(pos_x).to_value(u.m)) - r
    x_max = float(np.nanmax(pos_x).to_value(u.m)) + r
    y_min = float(np.nanmin(pos_y).to_value(u.m)) - r
    y_max = float(np.nanmax(pos_y).to_value(u.m)) + r
    bounds = PlotBounds(x_lim=(x_min, x_max), y_lim=(y_min, y_max))

    if axes_range:
        return patches, axes_range, highlighted_patches, bounds

    max_x = max(abs(x_min), abs(x_max))
    max_y = max(abs(y_min), abs(y_max))
    updated_axes_range = max(max_x, max_y) * 1.1

    return patches, updated_axes_range, highlighted_patches, bounds


@u.quantity_input(x=u.m, y=u.m, radius=u.m)
def get_telescope_patch(tel_type, x, y, radius, is_grayed_out=False):
    """
    Create patch for a telescope.

    Parameters
    ----------
    tel_type: str
        Telescope type.
    x : Quantity
        X position.
    y : Quantity
        Y position.
    radius : Quantity
        Telescope radius.
    is_grayed_out : bool
        Whether to plot telescope in gray.

    Returns
    -------
    patch : Patch
        Circle or rectangle patch.
    """
    config = leg_h.get_telescope_config(tel_type)
    x, y, r = x.to(u.m), y.to(u.m), radius.to(u.m)

    color = "gray" if is_grayed_out else config["color"]
    fill_flag = True if is_grayed_out else bool(config.get("filled", True))

    if config.get("shape", "circle") == "square":
        return mpatches.Rectangle(
            ((x - r / 2).value, (y - r / 2).value),
            width=r.value,
            height=r.value,
            fill=fill_flag,
            color=color,
        )
    if config.get("shape") == "hexagon":
        return mpatches.RegularPolygon(
            (x.value, y.value),
            numVertices=6,
            radius=r.value * np.sqrt(3) / 2,
            orientation=np.pi / 6,
            fill=fill_flag,
            color=color,
        )

    return mpatches.Circle(
        (x.value, y.value),
        radius=r.value,
        fill=fill_flag,
        alpha=0.5 if is_grayed_out else 1.0,
        color=color,
    )


def get_positions(telescopes):
    """
    Get X/Y positions depending on coordinate system.

    For ground coordinates, rotates the positions by 90 degrees.

    Returns
    -------
    x_rot, y_rot : Quantity
        Position coordinates.
    """
    if "position_x" in telescopes.colnames:
        x, y = telescopes["position_x"], telescopes["position_y"]
        locale_rotate_angle = 90 * u.deg
    elif "utm_east" in telescopes.colnames:
        x, y = telescopes["utm_east"], telescopes["utm_north"]
        locale_rotate_angle = 0 * u.deg
    else:
        raise ValueError("Missing required position columns.")

    return transf.rotate(x, y, locale_rotate_angle) if locale_rotate_angle != 0 else (x, y)


def create_patches(
    telescopes, scale, show_label, ax, grayed_out_elements=None, highlighted_elements=None
):
    """
    Create telescope patches and labels.

    Parameters
    ----------
    telescopes : Table
        Telescope data table.
    scale : float
        Marker size scale factor.
    show_label : bool
        Show telescope labels.
    ax : Axes
        Matplotlib axes object.
    grayed_out_elements : list or None
        List of telescope names to plot as gray circles.
    highlighted_elements : list or None
        List of telescope names to plot with red circles around them.

    Returns
    -------
    patches : list
        Shape patches.
    radii : list
        Telescope radii.
    highlighted_patches : list
        List of highlighted telescope patches.
    """
    patches, radii, highlighted_patches = [], [], []
    fontsize, scale_factor = (4, 2) if len(telescopes) > 30 else (8, 1)

    grayed_out_set = set(grayed_out_elements) if grayed_out_elements else set()
    highlighted_set = set(highlighted_elements) if highlighted_elements else set()

    for tel in telescopes:
        name = get_telescope_name(tel)
        radius = get_sphere_radius(tel)
        radii.append(radius)
        try:
            tel_type = names.get_array_element_type_from_name(name)
        except ValueError:
            tel_type = None

        is_grayed_out = name in grayed_out_set
        is_highlighted = name in highlighted_set

        patches.append(
            get_telescope_patch(
                tel_type,
                tel["pos_x_rotated"],
                tel["pos_y_rotated"],
                scale_factor * radius * scale,
                is_grayed_out=is_grayed_out,
            )
        )

        if is_highlighted:
            highlight_patch = mpatches.Circle(
                (tel["pos_x_rotated"].value, tel["pos_y_rotated"].value),
                radius=(scale_factor * radius * scale * 4).value,
                fill=False,
                color="red",
                linewidth=1,
            )
            highlighted_patches.append(highlight_patch)

        if show_label:
            ax.text(
                tel["pos_x_rotated"].value,
                tel["pos_y_rotated"].value + scale_factor * radius.value,
                name,
                ha="center",
                va="bottom",
                fontsize=fontsize * 0.8,
            )

    return patches, radii, highlighted_patches


def get_telescope_name(tel):
    """
    Get telescope name.

    Returns
    -------
    name : str
        Telescope name or fallback identifier.
    """
    if "telescope_name" in tel.colnames:
        return tel["telescope_name"]
    if "asset_code" in tel.colnames and "sequence_number" in tel.colnames:
        return f"{tel['asset_code']}-{tel['sequence_number']}"
    return f"tel_{tel.index}"


def get_sphere_radius(tel):
    """
    Get telescope sphere radius.

    Returns
    -------
    radius : Quantity
        Radius with units.
    """
    return tel["sphere_radius"] if "sphere_radius" in tel.colnames else 10.0 * u.m


def update_legend(ax, telescopes, grayed_out_elements=None, legend_location="best"):
    """Add legend for telescope types and counts."""
    grayed_out_set = set(grayed_out_elements) if grayed_out_elements else set()

    types = []
    for tel in telescopes:
        tel_name = get_telescope_name(tel)
        if tel_name not in grayed_out_set:
            types.append(names.get_array_element_type_from_name(tel_name))

    counts = Counter(types)

    objs, labels = [], []
    handler_map = {}

    for telescope_type in names.get_list_of_array_element_types():
        if counts[telescope_type]:
            objs.append(telescope_type)
            labels.append(f"{telescope_type} ({counts[telescope_type]})")

            class BaseLegendHandlerWrapper:  # pylint: disable=R0903
                """Wrapper for BaseLegendHandler to use in legend."""

                def __init__(self, tel_type):
                    self.tel_type = tel_type

                def legend_artist(self, legend, orig_handle, fontsize, handlebox):
                    handler = leg_h.BaseLegendHandler(self.tel_type)
                    return handler.legend_artist(legend, orig_handle, fontsize, handlebox)

            handler_map[telescope_type] = BaseLegendHandlerWrapper(telescope_type)

    ax.legend(objs, labels, handler_map=handler_map, prop={"size": 11}, loc=legend_location)


def finalize_plot(
    ax,
    patches,
    x_title,
    y_title,
    x_lim=None,
    y_lim=None,
    highlighted_patches=None,
):
    """Finalize plot appearance and limits."""
    ax.add_collection(PatchCollection(patches, match_original=True))

    if highlighted_patches:
        ax.add_collection(PatchCollection(highlighted_patches, match_original=True))

    ax.set(xlabel=x_title, ylabel=y_title)
    ax.tick_params(labelsize=8)
    ax.axis("square")
    if x_lim is not None and y_lim is not None:
        ax.set_xlim(*x_lim)
        ax.set_ylim(*y_lim)
    plt.tight_layout()



#!/usr/bin/python3

r"""
Plot simulated events.

Produces figures from one or more sim_telarray (.simtel.zst) files
It is meant to run after simulations (e.g., simtools-simulate-flasher,
simtools-simulate-illuminator).

What it does
------------
- Loads each provided sim_telarray file
- Generates selected plots (camera image, time traces, waveform matrices, peak timing, etc.)
- Optionally saves all figures to a single multi-page PDF per input file
- Optionally also saves individual PNGs

Command line arguments
----------------------
simtel_files (list, required)
    One or more sim_telarray files to visualize (.simtel.zst).
plots (list, optional)
    Which plots to generate. Choose from: event_image, time_traces, waveform_matrix,
    step_traces, integrated_signal_image, integrated_pedestal_image, peak_timing, all.
    Default: event_image.
tel_id (int, optional)
    Telescope ID to visualize. If omitted, the first available telescope will be used.
n_pixels (int, optional)
    For time_traces: number of pixel traces to draw. Default: 3.
pixel_step (int, optional)
    For step_traces and waveform_matrix: step between pixel indices. Default: 100.
max_pixels (int, optional)
    For step_traces: cap the number of plotted pixels. Default: None.
vmax (float, optional)
    For waveform_matrix: upper limit of color scale. Default: None.
half_width (int, optional)
    For integrated_*_image: half window width in samples. Default: 8.
offset (int, optional)
    For integrated_pedestal_image: offset between pedestal and peak windows. Default: 16.
sum_threshold (float, optional)
    For peak_timing: minimum pixel sum to consider a pixel. Default: 10.0.
peak_width (int, optional)
    For peak_timing: expected peak width in samples. Default: 8.
examples (int, optional)
    For peak_timing: show example traces. Default: 3.
timing_bins (int, optional)
    For peak_timing: number of histogram bins for peak sample. Default: None (contiguous bins).
distance (float, optional)
    Optional distance annotation for event_image.
output_file (str, optional)
    Base name for output. If provided, outputs will be placed under the standard IOHandler
    output directory and named ``<base>_<inputstem>.pdf``. If omitted, defaults are derived
    from each input file name.
save_pngs (flag, optional)
    Also save individual PNG files per figure.
dpi (int, optional)
    DPI for PNG outputs. Default: 300.
output_path (str, optional)
    Path to save the output files.

Examples
--------
1) Camera image and time traces for a single file, save a PDF:

   simtools-plot-simtel-events \
     --simtel_files tests/resources/ff-1m_flasher.simtel.zst \
     --plots event_image time_traces \
     --tel_id 1 \
     --output_file simulate_illuminator_inspect

2) All plots for multiple files, PNGs and PDFs:

   simtools-plot-simtel-events \
     --simtel_files f1.simtel.zst f2.simtel.zst \
     --plots all \
     --save_pngs --dpi 200

"""

from pathlib import Path

import simtools.utils.general as gen
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.visualization.plot_simtel_events import PLOT_CHOICES, generate_and_save_plots


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Create diagnostic plots from sim_telarray files using simtools visualization."
        ),
    )

    config.parser.add_argument(
        "--simtel_files",
        help="One or more sim_telarray files (.simtel.zst)",
        nargs="+",
        required=True,
    )
    config.parser.add_argument(
        "--plots",
        help=f"Plots to generate. Choices: {', '.join(sorted(PLOT_CHOICES))}",
        nargs="+",
        default=["event_image"],
        choices=sorted(PLOT_CHOICES),
    )
    config.parser.add_argument("--tel_id", type=int, default=None, help="Telescope ID")
    config.parser.add_argument(
        "--n_pixels", type=int, default=3, help="For time_traces: number of pixel traces"
    )
    config.parser.add_argument(
        "--pixel_step", type=int, default=100, help="Step between pixel ids for step plots"
    )
    config.parser.add_argument(
        "--max_pixels", type=int, default=None, help="Cap number of pixels for step traces"
    )
    config.parser.add_argument("--vmax", type=float, default=None, help="Color scale vmax")
    config.parser.add_argument(
        "--half_width", type=int, default=8, help="Half window width for integrated images"
    )
    config.parser.add_argument(
        "--offset",
        type=int,
        default=16,
        help="offset between pedestal and peak windows (integrated_pedestal_image)",
    )
    config.parser.add_argument(
        "--sum_threshold",
        type=float,
        default=10.0,
        help="Minimum pixel sum to consider in peak timing",
    )
    config.parser.add_argument(
        "--peak_width", type=int, default=8, help="Expected peak width in samples"
    )
    config.parser.add_argument(
        "--examples", type=int, default=3, help="Number of example traces to draw"
    )
    config.parser.add_argument(
        "--timing_bins",
        type=int,
        default=None,
        help="Number of bins for timing histogram (contiguous if not set)",
    )
    config.parser.add_argument(
        "--distance",
        type=float,
        default=None,
        help="Optional distance annotation for event_image (same units as input expects)",
    )
    config.parser.add_argument(
        "--event_index",
        type=int,
        default=None,
        help="0-based index of the event to plot; default is the first event",
    )
    config.parser.add_argument(
        "--output_file",
        type=str,
        default=None,
        help=(
            "Base name for output. If set, PDFs will be named '<base>_<inputstem>.pdf' "
            "in the standard output directory"
        ),
    )
    config.parser.add_argument(
        "--save_pngs",
        action="store_true",
        help="Also save individual PNG images per plot",
    )
    config.parser.add_argument("--dpi", type=int, default=300, help="PNG dpi")

    return config.initialize(db_config=False, require_command_line=True)


def main():
    """Generate plots from sim_telarray files."""
    app_context = startup_application(_parse)

    simtel_files = [
        Path(p).expanduser() for p in gen.ensure_iterable(app_context.args["simtel_files"])
    ]
    plots = list(gen.ensure_iterable(app_context.args.get("plots")))

    generate_and_save_plots(
        simtel_files=simtel_files, plots=plots, args=app_context.args, ioh=app_context.io_handler
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3

r"""
    Derive single photon electron spectrum from a given amplitude spectrum.

    Normalizes singe-p.e. amplitude distribution to mean amplitude of 1.0,
    as required by sim_telarray. Allows to fold in afterpulse distribution
    to prompt a spectrum. Uses the sim_telarray tool 'norm_spe' to normalize
    the spectra.

    Input files can be in ecsv format (preferred) or in the sim_telarray legacy format.

    Two output files with identical data are written to the output directory:

    - 'output_file'.ecsv: Single photon electron spectrum in ecsv format (data and metadata).
    - 'output_file'.dat: Single photon electron spectrum in sim_telarray format.

    Example
    -------

    .. code-block:: console

        simtools-derive-photon-electron-spectrum \\
            --input_spectrum spectrum_photon_electron.ecsv \\
            --afterpulse_spectrum spectrum_afterpulse.ecsv \\
            --step_size 0.02 \\
            --max_amplitude 42.0 \\
            --use_norm_spe \\
            --output_path ./tests/output \\
            --output_file spectrum_photon_electron_afterpulse.ecsv

    For an example of how to plot the single photon electron spectrum, see the
    integration test 'tests/integration_tests/config/plot_tabular_data_for_single_pe_data.yml'.

"""

from pathlib import Path

from simtools.application_control import get_application_label, startup_application
from simtools.camera.single_photon_electron_spectrum import SinglePhotonElectronSpectrum
from simtools.configuration import configurator


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Derive single photon electron spectrum from a given amplitude spectrum.",
    )
    config.parser.add_argument(
        "--input_spectrum",
        help="File with amplitude spectrum.",
        type=Path,
        required=True,
    )
    config.parser.add_argument(
        "--afterpulse_spectrum",
        help="File with afterpulse spectrum.",
        type=Path,
        required=False,
    )
    config.parser.add_argument(
        "--step_size",
        help="Step size in amplitude spectrum",
        type=float,
        default=0.02,
        required=False,
    )
    config.parser.add_argument(
        "--max_amplitude",
        help="Maximum amplitude for single p.e. for amplitude spectrum",
        type=float,
        default=42.0,
        required=False,
    )
    config.parser.add_argument(
        "--scale_afterpulse_spectrum",
        help="Scale afterpulse spectrum by the given factor",
        type=float,
        default=1.0,
        required=False,
    )
    config.parser.add_argument(
        "--afterpulse_amplitude_range",
        help="Amplitude range in pe for afterpulse calculation",
        type=float,
        nargs=2,
        default=[0.0, 42.0],
        required=False,
    )
    config.parser.add_argument(
        "--fit_afterpulse",
        help="Fit afterpulse spectrum with an exponential decay function.",
        action="store_true",
        required=False,
    )
    config.parser.add_argument(
        "--afterpulse_decay_factor_fixed_value",
        help="Fix decay factor in afterpulse fit (free fit parameter if not set set).",
        type=float,
        default=15.0,
        required=False,
    )
    config.parser.add_argument(
        "--use_norm_spe",
        help="Use sim_telarray tool 'norm_spe' to normalize the spectrum.",
        action="store_true",
        required=False,
    )

    return config.initialize(db_config=False, output=True, simulation_model=["telescope"])


def main():
    """Derive single photon electron spectrum from a given amplitude spectrum."""
    app_context = startup_application(_parse)

    single_pe = SinglePhotonElectronSpectrum(app_context.args)
    single_pe.derive_single_pe_spectrum()
    single_pe.write_single_pe_spectrum()


if __name__ == "__main__":
    main()
